/*
 * eth_ecat_pdo_promise.hpp
 *
 *  Created on: Jun 1, 2023
 *      Author: algin
 */

#ifndef FREE_RTOS_ETHERNET_INDUSTRY_COE_ETH_ECAT_PDO_PROMISE_HPP_
#define FREE_RTOS_ETHERNET_INDUSTRY_COE_ETH_ECAT_PDO_PROMISE_HPP_

#include <cstdint>

#include "semaphore/semaphore.hpp"

#include "ethernet_industry/eth_ecat_command.hpp"
#include "ethernet_industry/eth_ecat_custom_tuple.hpp"

namespace free_rtos {

namespace pdo_promise {

template<typename... DataTypes>
class PDOFuture {
public:
    PDOFuture(DataTypes&... data)
    : data_tuple_{data...} { }

    bool is_ready() {
        return ready_;
    }

    custom_tuple<DataTypes&...> get() {
        ready_ = false;
        sem_.pend();

        return data_tuple_;
    }

    void pack(uint8_t* raw) {
        PackFunctor functor{raw};

        for_each(data_tuple_, functor);

        sem_.post();
        ready_ = true;
    }

    void unpack(uint8_t* raw) {
        UnpackFunctor functor{raw};

        for_each(data_tuple_, functor);

        sem_.post();
        ready_ = true;
    }

private:
    custom_tuple<DataTypes&...> data_tuple_;
    free_rtos::Semaphore sem_;
    bool ready_{false};
};

class IPDOPromise {
public:
    IPDOPromise(address::Offset offset)
    : offset_{offset} { }

    IPDOPromise* get_next() {
        return next_;
    }

    IPDOPromise& operator+(IPDOPromise &next) {
        next_ = &next;

        return next;
    }

    virtual void set_value(uint8_t* process_data, uint32_t len) = 0;

protected:
    address::Offset offset_;

private:
    IPDOPromise *next_{nullptr};
};

template<typename... DataTypes>
class PDOWritePromise : public IPDOPromise {
public:
    PDOWritePromise(address::Offset offset, DataTypes&... data)
    : IPDOPromise{offset}
    , future_{data...} { }

    PDOFuture<DataTypes...>& get_future() {
        return future_;
    }

    virtual void set_value(uint8_t* process_data, uint32_t len) override {
        future_.pack(process_data + offset_);
    }

private:
    PDOFuture<DataTypes...> future_;
};

template<typename... DataTypes>
class PDOReadPromise : public IPDOPromise {
public:
    PDOReadPromise(address::Offset offset, DataTypes&... data)
    : IPDOPromise{offset}
    , future_{data...} { }

    PDOFuture<DataTypes...>& get_future() {
        return future_;
    }

    virtual void set_value(uint8_t* process_data, uint32_t len) override {
        future_.unpack(process_data + offset_);
    }

private:
    PDOFuture<DataTypes...> future_;
};

} // namespace pdo_promise

} // namespace free_rtos

#endif /* FREE_RTOS_ETHERNET_INDUSTRY_COE_ETH_ECAT_PDO_PROMISE_HPP_ */
